from .iolayers import ReadinLayer as ReadinLayer
from .iolayers import ReadoutLayer as ReadoutLayer
from .spectralconv import SpectralConv1DLayer as SpectralConv1DLayer
from .spectralconv import SpectralConv2DLayer as SpectralConv2DLayer
from .spectralconv import SpectralConv3DLayer as SpectralConv3DLayer




from torch.types import Tensor
from typing import Tuple, Union, Callable

import torch

from ..base import NeuralOperator
from ..layers.iolayers import ReadinLayer, ReadoutLayer
from ..layers.spectralconv import SpectralConv1DLayer, SpectralConv2DLayer, SpectralConv3DLayer


class FourierOperator(NeuralOperator):
    """
    Fourier Neural Operator. 
    It learns a spectral kernel that can be used to approximate functions in the frequency domain.
    """
    
    def __init__(self, 
                 in_features: int, 
                 hidden_features: int, 
                 out_features: int,
                 modes: Union[int, Tuple], 
                 depth: int = 3,
                 activation_function: Callable[[Tensor], Tensor] = torch.relu
                 ):
        """
        Initializes the Fourier operator with the given parameters.
        Args:
            in_features (int): Number of input features.
            hidden_features (int): Number of hidden features.
            out_features (int): Number of output features.
            modes (Union[int, Tuple]): Number of Fourier modes to consider. 
                                       If a tuple, it should contain the number of modes for each dimension.
            depth (int): Depth of the operator, i.e., number of spectral convolution layers.
            activation_function (Callable[[Tensor], Tensor]): Activation function to apply after each layer.
        """
        super().__init__()
        self.readin = ReadinLayer(in_features, hidden_features)

        if isinstance(modes, tuple) and len(modes) > 3: 
            raise NotImplementedError("FourierOperator currently supports up to 3D inputs.")
        
        if isinstance(modes, int):
            self.kernel_integral = torch.nn.ModuleList([
                SpectralConv1DLayer(hidden_features, hidden_features, modes = modes) for _ in range(depth)
             ])
        elif len(modes) == 2:
            self.kernel_integral = torch.nn.ModuleList([
                SpectralConv2DLayer(hidden_features, hidden_features, mode_h = modes[0], mode_w = modes[1]) for _ in range(depth)
            ])
        elif len(modes) == 3:
            self.kernel_integral = torch.nn.ModuleList([
                SpectralConv3DLayer(hidden_features, hidden_features, mode_d = modes[0], mode_h = modes[1], mode_w = modes[2]) for _ in range(depth)
            ])
        
        self.readout = ReadoutLayer(hidden_features, out_features)
        self.depth = depth
        self.activation_function = activation_function

    def forward(self, x) -> Tensor:
        """
        Forward pass for the Fourier operator.
        
        Args:
            x (Tensor): Input tensor.
        
        Returns:
            Tensor: Tensor after applying the Fourier operator.
        """
        x = self.readin(x)
        x = self.activation_function(x)

        for layer in self.kernel_integral:
            x = layer(x)
            x = self.activation_function(x)

        x = self.readout(x)
        return x


r""" 
.. include:: ../README.md
""" 

__version__ = "0.0.4"
__authors__ = [
    "Hrishikesh Belagali",
    "Aditya Narayan"
]
__author_emails__ = [
    "belagal1@msu.edu",
    "ma24btech11001@iith.ac.in"
]
__url__ = "https://github.com/lonelyneutrin0/neurop"

from .base import NeuralOperator as NeuralOperator

from . import operators as operators
from . import layers as layers



from abc import ABC, abstractmethod

import torch 
from torch.types import Tensor
from typing import TypedDict


class Metrics(TypedDict, total=False):
    """
    TypedDict to hold metrics for evaluation.
    All fields are optional for flexibility
    """
    loss: float
    accuracy: float
class NeuralOperator(torch.nn.Module, ABC):
    """
    Abstract class for Neural Operators.

    Attributes: 
        readin (torch.nn.Module):
            Reads in input data and projects to higher dimensional space 
        kernel_integral (torch.nn.Module):
            The kernel integral operator that performs the main computation
        readout (torch.nn.Module):
            Reads out data to lower dimensional space
        optimizer (torch.optim.Optimizer):
            Optimization algorithm to choose. Defaults to Adam(lr=1e-3)
        activation_function (Callable[[Tensor], Tensor]):
            Activation to introduce nonlinearity between kernel operations
    """

    def __init__(self,):
        """
        __init__ method to initialize NeuralOperator
        """
        super().__init__()
        pass
    
    @abstractmethod
    def forward(self, x: Tensor) -> Tensor:
        """
        Forward pass to be implemented by subclasses.
        """
        pass

    # @abstractmethod
    # def loss(self, prediction: Tensor, target: Tensor) -> Tensor:
    #     """
    #     Loss function specific to the problem/operator.
    #     """
    #     pass

    # @abstractmethod
    # def train_step(self, x: Tensor, y: Tensor) -> Tensor:
    #     """
    #     One training step: forward + loss + backward + optimizer step.
    #     """
    #     pass

    # @abstractmethod
    # def evaluate(self, x: Tensor, y: Tensor) -> float:
    #     """
    #     Evaluate model performance on validation/test data.
    #     """
    #     pass

    # @abstractmethod
    # def save(self, path: Path): 
    #     """
    #     Write model parameters to a file  
    #     """
    #     pass

    # @classmethod
    # @abstractmethod
    # def load(cls, path: Path):
    #     """
    #     Load a neural operator from a file
    #     """
    #     pass

    # def to_device(self, device: torch.device): 
    #     """
    #     Send data to a Torch device 
    #     """
    #     self.readin.to(device)
    #     self.kernel_integral.to(device)
    #     self.readout.to(device)

    # @abstractmethod
    # def calculate_metrics(self, ground_truth: Tensor, predicted: Tensor) -> Metrics: 
    #     """
    #     Compute the desired metrics and output a TypedDict 
    #     """
    #     pass

    # @abstractmethod
    # def __repr__(self) -> str:
    #     """
    #     Brief overview of the model architecture
    #     """
    #     pass


import torch

from torch import Tensor


class SpectralConv1DLayer(torch.nn.Module):
    """
    Spectral Convolution 1 Dimensional Layer

    The input is assumed to have shape (B, C, L) where:
        - B is the batch size,
        - C is the number of input channels,
        - L is the length of the signal,
    
    The layer transforms the input to the frequency domain using a 1D FFT along the last dimension,
    $$x_f = \mathcal{F}[x]$$ 
    Then, the layer applies a pointwise multiplication (which is equivalent to a convolution in the time domain) as follows: 
    $$y_{b, o, k} = \sum_{c=0}^{C-1} W_{c, o, k} x_{b, c, k}$$
    where the contraction is along the features. Finally, the output is transformed back to the time domain using the inverse FFT:
    $$y = \mathcal{F}^{-1}[y]$$
    
    Args:
        in_features (int): Number of input channels.
        out_features (int): Number of output channels.
        modes (int): Number of Fourier modes to consider.
        init_scale (float): Initialization scale for weights.
    """

    def __init__(self, in_features: int, out_features: int, modes: int, init_scale: float = 1.0):
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.modes = modes
        
        scale = init_scale / (in_features * out_features)
        self.weight = torch.nn.Parameter(
            torch.randn(in_features, out_features, modes, dtype=torch.cfloat) * scale
        )

    def forward(self, x: Tensor) -> Tensor:
        """
        Forward pass for the spectral convolution layer.
        
        Args:
            x (Tensor): Input tensor of shape (B, C, L)
        
        Returns:
            Tensor: Output tensor of shape (B, out_features, L)
        """
        batchsize, c, length = x.shape
        
        # FFTs only produce n//2 + 1 nonredundant modes for real inputs
        modes = min(self.modes, length // 2 + 1)  
        
        x_ft = torch.fft.rfft(x, dim=2, norm='ortho')
        
        out_ft = torch.zeros(
            batchsize, self.out_features, x_ft.size(2), 
            dtype=torch.cfloat, device=x.device
        )
        
        if modes > 0:
            out_ft[:, :, :modes] = torch.einsum(
                "bck,cok->bok", 
                x_ft[:, :, :modes], 
                self.weight[:, :, :modes]
            )
        
        x_out = torch.fft.irfft(out_ft, n=length, dim=2, norm='ortho')
        return x_out

class SpectralConv2DLayer(torch.nn.Module):
    """
    Spectral Convolution 2 Dimensional Layer    
    The input is assumed to have shape (B, C, H, W) where:
        - B is the batch size,
        - C is the number of input channels,
        - H is the height of the signal,
        - W is the width of the signal,
    The layer transforms the input to the frequency domain using a 2D FFT along the last two dimensions,
    $$x_f = \mathcal{F}[x]$$
    Then, the layer applies a pointwise multiplication (which is equivalent to a convolution in the time domain) as follows:
    $$y_{b, o, k_h, k_w} = \sum_{c=0}^{C-1} W_{c, o, k_h, k_w} x_{b, c, k_h, k_w}$$
    where the contraction is along the features. Finally, the output is transformed back to the time domain using the inverse FFT:
    $$y = \mathcal{F}^{-1}[y]$$
    """

    def __init__(self, in_features: int, out_features: int, mode_h: int, mode_w: int, init_scale: float = 1.0):
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.mode_h = mode_h
        self.mode_w = mode_w
        
        scale = init_scale / (in_features * out_features)
        self.weight = torch.nn.Parameter(
            torch.randn(in_features, out_features, mode_h, mode_w, dtype=torch.cfloat) * scale
        )

    def forward(self, x: Tensor) -> Tensor:
        """
        Forward pass for the Spectral Convolution 2D layer.
        
        Args:
            x (Tensor): Input tensor of shape (B, C, H, W) - PyTorch convention
        
        Returns:
            Tensor: Output tensor of shape (B, out_features, H, W)
        """
        batchsize, c, h, w = x.shape
        
        mode_h = min(self.mode_h, h)
        mode_w = min(self.mode_w, w // 2 + 1) 
        
        x_ft = torch.fft.rfft2(x, dim=(2, 3), norm='ortho')
        
        out_ft = torch.zeros(
            batchsize, self.out_features, h, x_ft.size(3),
            dtype=torch.cfloat, device=x.device
        )
        
        if mode_h > 0 and mode_w > 0:
            out_ft[:, :, :mode_h, :mode_w] = torch.einsum(
                "bchw,cohw->bohw", 
                x_ft[:, :, :mode_h, :mode_w], 
                self.weight[:, :, :mode_h, :mode_w]
            )
        
        x_out = torch.fft.irfft2(out_ft, s=(h, w), dim=(2, 3), norm='ortho')
        return x_out

class SpectralConv3DLayer(torch.nn.Module):
    """
    Spectral Convolution 3 Dimensional Layer
    The input is assumed to have shape (B, C, D, H, W) where:
        - B is the batch size,
        - C is the number of input channels,
        - D is the depth of the signal,
        - H is the height of the signal,
        - W is the width of the signal,
    The layer transforms the input to the frequency domain using a 3D FFT along the last three dimensions,
        $$x_f = \mathcal{F}[x]$$
    Then, the layer applies a pointwise multiplication (which is equivalent to a convolution in the time domain) as follows:
        $$y_{b, o, k_d, k_h, k_w} = \sum_{c=0}^{C-1} W_{c, o, k_d, k_h, k_w} x_{b, c, k_d, k_h, k_w}$$
    where the contraction is along the features. Finally, the output is transformed back to the time domain using the inverse FFT:
        $$y = \mathcal{F}^{-1}[y]$$
    """

    def __init__(self, in_features: int, out_features: int, mode_d: int, mode_h: int, mode_w: int, init_scale: float = 1.0):
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.mode_d = mode_d
        self.mode_h = mode_h
        self.mode_w = mode_w
        
        scale = init_scale / (in_features * out_features)
        self.weight = torch.nn.Parameter(
            torch.randn(in_features, out_features, mode_d, mode_h, mode_w, dtype=torch.cfloat) * scale
        )

    def forward(self, x: Tensor) -> Tensor:
        """
        Forward pass for the Spectral Convolution 3D layer.
        Args:
            x (Tensor): Input tensor of shape (B, C, D, H, W)
        Returns:
            Tensor: Output tensor of shape (B, out_features, D, H, W)
        """
        batchsize, c, d, h, w = x.shape
        
        mode_d = min(self.mode_d, d)
        mode_h = min(self.mode_h, h)
        mode_w = min(self.mode_w, w // 2 + 1)

        x_ft = torch.fft.rfftn(x, dim=(2, 3, 4), norm='ortho')

        out_ft = torch.zeros(
            batchsize, self.out_features, d, h, x_ft.size(4),
            dtype=torch.cfloat, device=x.device
        )

        if mode_d > 0 and mode_h > 0 and mode_w > 0:
            out_ft[:, :, :mode_d, :mode_h, :mode_w] = torch.einsum(
                "bcdhw,codhw->bohdw", 
                x_ft[:, :, :mode_d, :mode_h, :mode_w], 
                self.weight[:, :, :mode_d, :mode_h, :mode_w]
            )
        
        x_out = torch.fft.irfftn(out_ft, s=(d, h, w), dim=(2, 3, 4), norm='ortho')
        return x_out


import torch
from torch.types import Tensor
class ReadinLayer(torch.nn.Module):
    """
    Reads in input data and projects it to a higher dimensional space.
    """
    
    def __init__(self, in_features: int, hidden_features: int):
        """
        Args:   
            in_features (int): Number of features of input data.
            hidden_features (int): Number of hidden featuers of transformed data.
        """
        super().__init__()
        self.linear = torch.nn.Linear(in_features, hidden_features)
    
    def forward(self, x: Tensor) -> Tensor:
        """
        Args:
            x (Tensor): Input data of shape (batch_size, in_features, d_1, d_2, .... d_n).
        """ 

        # Reshape input to (batch_size, d_1, d_2, .... d_n, in_features) 
        x = x.permute(0, *range(2, x.ndim), 1)
        y = self.linear(x)
    
        # Reshape back to (batch_size, hidden_features, d_1, d_2, .... d_n)
        return y.permute(0, -1, *range(1, y.ndim - 1))  
class ReadoutLayer(torch.nn.Module):
    """
    Reads out data to lower dimensional space.
    """
    
    def __init__(self, hidden_features: int, output_features: int):
        """
        Args:   
            hidden_features (int): Dimension of the input data.
            output_features (int): Dimension of the output data after projection.
        """
        super().__init__()
        self.linear = torch.nn.Linear(hidden_features, output_features)
    
    def forward(self, x: Tensor) -> Tensor:
        """"
        Args:
            x (Tensor): Input data of shape (batch_size, hidden_features, d_1, d_2, .... d_n).
        """

        # Reshape input to (batch_size, d_1, d_2, .... d_n, hidden_features)
        x = x.permute(0, *range(2, x.ndim), 1)
        y = self.linear(x)

        # Reshape back to (batch_size, output_features, d_1, d_2, .... d_n)
        return y.permute(0, -1, *range(1, y.ndim - 1))
    